@@ -32,6 +32,7 @@
 #include "image.h"
 #include "text.h"
 #include "link.h"
+#include "table.h"
 
 static void
 undo_cb (BonoboUIComponent *uic, GtkHTMLControlData *cd, const char *cname)
@@ -66,13 +67,13 @@ paste_cb (BonoboUIComponent *uic, GtkHTMLControlData *cd, const char *cname)
 static void
 search_cb (BonoboUIComponent *uic, GtkHTMLControlData *cd, const char *cname)
 {
-	search (cd, TRUE);
+	search (cd, FALSE);
 }
 
 static void
 search_regex_cb (BonoboUIComponent *uic, GtkHTMLControlData *cd, const char *cname)
 {
-	search (cd, FALSE);
+	search (cd, TRUE);
 }
 
 static void
@@ -81,6 +82,12 @@ search_next_cb (BonoboUIComponent *uic, GtkHTMLControlData *cd, const char *cnam
 	search_next (cd);
 }
 
+static void
+select_all_cb (BonoboUIComponent *uic, GtkHTMLControlData *cd, const char *cname)
+{
+	gtk_html_command (cd->html, "select-all");
+}
+
 static void
 replace_cb (BonoboUIComponent *uic, GtkHTMLControlData *cd, const char *cname)
 {
@@ -146,6 +153,23 @@ insert_rule_cb (BonoboUIComponent *uic, GtkHTMLControlData *cd, const char *cnam
 	gtk_html_edit_properties_dialog_show (cd->properties_dialog);
 }
 
+static void
+insert_table_cb (BonoboUIComponent *uic, GtkHTMLControlData *cd, const char *cname)
+{
+	if (cd->properties_dialog)
+		gtk_html_edit_properties_dialog_close (cd->properties_dialog);
+
+	cd->properties_dialog = gtk_html_edit_properties_dialog_new (cd, TRUE, _("Insert"));
+
+	gtk_html_edit_properties_dialog_add_entry (cd->properties_dialog,
+						   GTK_HTML_EDIT_PROPERTY_TABLE, _("Table"),
+						   table_insert,
+						   table_insert_cb,
+						   table_close_cb);
+
+	gtk_html_edit_properties_dialog_show (cd->properties_dialog);
+}
+
 static void
 properties_cb (BonoboUIComponent *uic, GtkHTMLControlData *cd, const char *cname)
 {
@@ -161,14 +185,16 @@ BonoboUIVerb verbs [] = {
 	BONOBO_UI_UNSAFE_VERB ("EditCopy", copy_cb),
 	BONOBO_UI_UNSAFE_VERB ("EditPaste", paste_cb),
 	BONOBO_UI_UNSAFE_VERB ("EditFind", search_cb),
-	BONOBO_UI_UNSAFE_VERB ("EditFindRegexp", search_regex_cb),
+	BONOBO_UI_UNSAFE_VERB ("EditFindRegex", search_regex_cb),
 	BONOBO_UI_UNSAFE_VERB ("EditFindAgain", search_next_cb),
 	BONOBO_UI_UNSAFE_VERB ("EditReplace", replace_cb),
 	BONOBO_UI_UNSAFE_VERB ("EditProperties", properties_cb),
+	BONOBO_UI_UNSAFE_VERB ("EditSelectAll", select_all_cb),
 
 	BONOBO_UI_UNSAFE_VERB ("InsertImage", insert_image_cb),
-	BONOBO_UI_UNSAFE_VERB ("InsertLink", insert_link_cb),
-	BONOBO_UI_UNSAFE_VERB ("InsertRule", insert_rule_cb),
+	BONOBO_UI_UNSAFE_VERB ("InsertLink",  insert_link_cb),
+	BONOBO_UI_UNSAFE_VERB ("InsertRule",  insert_rule_cb),
+	BONOBO_UI_UNSAFE_VERB ("InsertTable", insert_table_cb),
 
 	BONOBO_UI_VERB_END
 };
