@@ -29,6 +29,8 @@
 
 #include <gnome.h>
 
+#include "htmlcolorset.h"
+#include "htmlcursor.h"
 #include "htmlengine-edit.h"
 #include "htmlengine-edit-clueflowstyle.h"
 #include "htmlengine-edit-cut-and-paste.h"
@@ -37,14 +39,18 @@
 #include "htmlengine-edit-movement.h"
 #include "htmlengine-edit-cursor.h"
 #include "htmlengine-edit-table.h"
+#include "htmlengine-edit-tablecell.h"
 #include "htmlengine-edit-text.h"
 #include "htmlengine-edit-selection-updater.h"
 #include "htmlengine-print.h"
 #include "htmlengine-save.h"
+#include "htmlframe.h"
+#include "htmliframe.h"
+#include "htmlimage.h"
+#include "htmlplainpainter.h"
 #include "htmlsettings.h"
-#include "htmlcolorset.h"
+#include "htmltable.h"
 #include "htmlselection.h"
-#include "htmlcursor.h"
 
 #include "gtkhtml.h"
 #include "gtkhtml-embedded.h"
@@ -93,13 +99,13 @@ enum {
 static guint signals [LAST_SIGNAL] = { 0 };
 
 /* keybindings signal hadlers */
-static void    scroll                 (GtkHTML *html, GtkOrientation orientation, GtkScrollType scroll_type, gfloat position);
-static void    cursor_move            (GtkHTML *html, GtkDirectionType dir_type, GtkHTMLCursorSkipType skip);
-static void    command                (GtkHTML *html, GtkHTMLCommandType com_type);
-static gint    mouse_change_pos       (GtkWidget *widget, GdkWindow *window, gint x, gint y);
-static void    load_keybindings       (GtkHTMLClass *klass);
-static void    set_editor_keybindings (GtkHTML *html, gboolean editable);
-static gchar * get_value_nick         (GtkHTMLCommandType com_type);
+static void     scroll                 (GtkHTML *html, GtkOrientation orientation, GtkScrollType scroll_type, gfloat position);
+static void     cursor_move            (GtkHTML *html, GtkDirectionType dir_type, GtkHTMLCursorSkipType skip);
+static gboolean command                (GtkHTML *html, GtkHTMLCommandType com_type);
+static gint     mouse_change_pos       (GtkWidget *widget, GdkWindow *window, gint x, gint y);
+static void     load_keybindings       (GtkHTMLClass *klass);
+static void     set_editor_keybindings (GtkHTML *html, gboolean editable);
+static gchar *  get_value_nick         (GtkHTMLCommandType com_type);
 
 
 /* Values for selection information.  FIXME: what about COMPOUND_STRING and
@@ -646,9 +652,9 @@ key_press_event (GtkWidget *widget,
 
 		/* printf ("event length: %d s[0]: %d string: '%s'\n", event->length, event->string [0], event->string); */
 		str = e_utf8_from_gtk_event_key (widget, event->keyval, event->string);
-		/* printf ("len: %d str: %s\n", str ? unicode_strlen (str, -1) : -1, str); */
+		/* printf ("len: %d str: %s\n", str ? g_utf8_strlen (str, -1) : -1, str); */
 		if (str)
-			html_engine_paste_text (html->engine, str, unicode_strlen (str, -1));
+			html_engine_paste_text (html->engine, str, g_utf8_strlen (str, -1));
 		else if (event->length == 1 && event->string
 			 && ((guchar)event->string [0]) > 0x20 && ((guchar)event->string [0]) < 0x80)
 			html_engine_paste_text (html->engine, event->string, 1);
@@ -688,7 +694,8 @@ realize (GtkWidget *widget)
 			       (gdk_window_get_events (html->layout.bin_window)
 				| GDK_EXPOSURE_MASK | GDK_POINTER_MOTION_MASK
 				| GDK_ENTER_NOTIFY_MASK
-				| GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK
+				| GDK_BUTTON_PRESS_MASK 
+				| GDK_BUTTON_RELEASE_MASK
 				| GDK_KEY_PRESS_MASK | GDK_KEY_RELEASE_MASK));
 
 	html_engine_realize (html->engine, html->layout.bin_window);
@@ -769,9 +776,7 @@ size_allocate (GtkWidget *widget, GtkAllocation *allocation)
 		e->width  = allocation->width;
 		e->height = allocation->height;
 
-		if (e->clue && e->painter
-		    && html_object_calc_min_width (e->clue, e->painter) < e->width - e->leftBorder - e->rightBorder)
-			html_engine_calc_size (html->engine);
+		html_engine_calc_size (html->engine);
 
 		gtk_html_private_calc_scrollbars (html);
 	}
@@ -781,6 +786,21 @@ size_allocate (GtkWidget *widget, GtkAllocation *allocation)
 #endif
 }
 
+static void
+set_pointer_url (GtkHTML *html, const char *url)
+{
+	if (url == html->pointer_url)
+		return;
+
+	if (url && html->pointer_url && !strcmp (url, html->pointer_url))
+		return;
+		
+	g_free (html->pointer_url);
+	html->pointer_url = url ? g_strdup (url) : NULL;
+	gtk_signal_emit (GTK_OBJECT (html), 
+			 signals[ON_URL], html->pointer_url);
+}
+
 static void
 on_object (GtkWidget *widget, GdkWindow *window, HTMLObject *obj)
 {
@@ -788,35 +808,25 @@ on_object (GtkWidget *widget, GdkWindow *window, HTMLObject *obj)
 	const gchar *url;
 
 	if (obj) {
-		if ((url = (obj) ? html_object_get_url (obj) : NULL)) {
-			if (html->pointer_url == NULL || strcmp (html->pointer_url, url) != 0) {
-				g_free (html->pointer_url);
-				html->pointer_url = g_strdup (url);
-				gtk_signal_emit (GTK_OBJECT (html), signals[ON_URL], url);
-			}
-
+		url = html_object_get_url (obj);
+		if (url != NULL) {
+			set_pointer_url (html, url);
+			
 			if (html->engine->editable)
 				gdk_window_set_cursor (window, html->ibeam_cursor);
 			else
 				gdk_window_set_cursor (window, html->hand_cursor);
 		} else {
-			if (html->pointer_url != NULL) {
-				g_free (html->pointer_url);
-				html->pointer_url = NULL;
-				gtk_signal_emit (GTK_OBJECT (html), signals [ON_URL], NULL);
-			}
-
-			if (obj != NULL && html_object_is_text (obj) && html->allow_selection)
+			set_pointer_url (html, NULL);				
+			
+			if (html_object_is_text (obj) && html->allow_selection)
 				gdk_window_set_cursor (window, html->ibeam_cursor);
 			else
 				gdk_window_set_cursor (window, html->arrow_cursor);
 		}
 	} else {
-		if (html->pointer_url) {
-			g_free (html->pointer_url);
-			html->pointer_url = NULL;
-			gtk_signal_emit (GTK_OBJECT (html), signals [ON_URL], NULL);
-		}
+		set_pointer_url (html, NULL);
+
 		gdk_window_set_cursor (window, html->arrow_cursor);
 	}
 }
@@ -860,8 +870,12 @@ mouse_change_pos (GtkWidget *widget, GdkWindow *window, gint x, gint y)
 		}
 
 		if (HTML_DIST ((x + engine->x_offset - html->selection_x1),
-			  (y + engine->y_offset - html->selection_y1)) > 4)
+			       (y + engine->y_offset - html->selection_y1)) 
+		    > html_painter_get_space_width (engine->painter, 
+						    GTK_HTML_FONT_STYLE_SIZE_3,  
+						    NULL)) {
 			html->in_selection = TRUE;
+		}
 
 		need_scroll = FALSE;
 
@@ -891,8 +905,10 @@ mouse_change_pos (GtkWidget *widget, GdkWindow *window, gint x, gint y)
 		if (engine->mark == NULL && engine->editable)
 			html_engine_set_mark (engine);
 
-		html_engine_select_region (engine, html->selection_x1, html->selection_y1,
-					   x + engine->x_offset, y + engine->y_offset);
+		html_engine_select_region (engine, html->selection_x1,
+					   html->selection_y1,
+					   x + engine->x_offset, 
+					   y + engine->y_offset);
 	}
 
 	on_object (widget, window, obj);
@@ -972,22 +988,30 @@ button_press_event (GtkWidget *widget,
 		switch (event->button) {
 		case 4:
 			/* Mouse wheel scroll up.  */
-			value = vadj->value - vadj->step_increment * 3;
+			if (event->state & GDK_CONTROL_MASK)
+				gtk_html_command (html, "zoom-in");
+			else {
+				value = vadj->value - vadj->step_increment * 3;
 			
-			if (value < vadj->lower)
-				value = vadj->lower;
+				if (value < vadj->lower)
+					value = vadj->lower;
 			
-			gtk_adjustment_set_value (vadj, value);
+				gtk_adjustment_set_value (vadj, value);
+			}
 			return TRUE;
 			break;
 		case 5:
 			/* Mouse wheel scroll down.  */
-			value = vadj->value + vadj->step_increment * 3;
+			if (event->state & GDK_CONTROL_MASK) 
+				gtk_html_command (html, "zoom-out");
+			else {
+				value = vadj->value + vadj->step_increment * 3;
 			
-			if (value > (vadj->upper - vadj->page_size))
-				value = vadj->upper - vadj->page_size;
+				if (value > (vadj->upper - vadj->page_size))
+					value = vadj->upper - vadj->page_size;
 			
-			gtk_adjustment_set_value (vadj, value);
+				gtk_adjustment_set_value (vadj, value);
+			}
 			return TRUE;
 			break;
 		case 2:
@@ -1062,6 +1086,7 @@ button_release_event (GtkWidget *widget,
 
 	if (event->button == 1) {
 		html->button1_pressed = FALSE;
+		
 		if (html->pointer_url != NULL && ! html->in_selection)
 			gtk_signal_emit (GTK_OBJECT (widget), signals[LINK_CLICKED], html->pointer_url);
 	}
@@ -1134,6 +1159,37 @@ enter_notify_event (GtkWidget *widget, GdkEventCrossing *event)
 
 /* X11 selection support.  */
 
+static guchar *
+replace_nbsp (const guchar *text)
+{
+	guchar *nt, *ntext = g_new (guchar, strlen (text) + 1);
+	gint pos = 0;
+
+	nt = ntext;
+	while (*text) {
+		if (*text == 0xc2 && pos == 0)
+			pos = 1;
+		else if (*text == 0xa0 && pos == 1) {
+			*nt = ' ';
+			nt ++;
+			pos = 0;
+		} else {
+			if (pos) {
+				*nt = 0xc2;
+				nt ++;
+				pos = 0;
+			}
+			*nt = *text;
+			nt ++;
+		}
+		text ++;
+	}
+
+	*nt = 0;
+
+	return ntext;
+}
+
 static void
 selection_get (GtkWidget        *widget, 
 	       GtkSelectionData *selection_data,
@@ -1162,42 +1218,49 @@ selection_get (GtkWidget        *widget,
 	
 	if (selection_string != NULL) {
 		if (info == TARGET_UTF8_STRING) {
+			printf ("UTF8_STRING\n");
 			gtk_selection_data_set (selection_data,
 						gdk_atom_intern ("UTF8_STRING", FALSE), 8,
 						(const guchar *) selection_string,
 						strlen (selection_string));
 		} else if (info == TARGET_UTF8) {
+			printf ("UTF-8\n");
 			gtk_selection_data_set (selection_data,
 						gdk_atom_intern ("UTF-8", FALSE), 8,
 						(const guchar *) selection_string,
 						strlen (selection_string));
-		} else if (info == TARGET_STRING) {
+		} else if (info == TARGET_STRING || info == TARGET_TEXT || info == TARGET_COMPOUND_TEXT) {
+			gchar *to_be_freed;
+
+			to_be_freed = selection_string;
+			selection_string = replace_nbsp (selection_string);
+			g_free (to_be_freed);
 			localized_string = e_utf8_to_gtk_string (widget,
 								 selection_string);
+
+			if (info == TARGET_STRING) {
+				printf ("STRING\n");
+				gtk_selection_data_set (selection_data,
+							GDK_SELECTION_TYPE_STRING, 8,
+							(const guchar *) localized_string, 
+							strlen (localized_string));
+			} else {
+				guchar *text;
+				GdkAtom encoding;
+				gint format;
+				gint new_length;
 			
-			gtk_selection_data_set (selection_data,
-						GDK_SELECTION_TYPE_STRING, 8,
-						(const guchar *) localized_string, 
-						strlen (localized_string));
+				printf ("TEXT or COMPOUND_TEXT\n");
+				gdk_string_to_compound_text (localized_string, 
+							     &encoding, &format,
+							     &text, &new_length);
+
+				gtk_selection_data_set (selection_data,
+							encoding, format,
+							text, new_length);
+				gdk_free_compound_text (text);
+			}
 			
-		} else if ((info == TARGET_TEXT) 
-			   || (info == TARGET_COMPOUND_TEXT)) {
-			guchar *text;
-			GdkAtom encoding;
-			gint format;
-			gint new_length;
-			
-			localized_string = e_utf8_to_gtk_string (widget,
-								 selection_string);
-			
-			gdk_string_to_compound_text (localized_string, 
-						     &encoding, &format,
-						     &text, &new_length);
-			
-			gtk_selection_data_set (selection_data,
-						encoding, format,
-						text, new_length);
-			gdk_free_compound_text (text);
 		}
 		g_free (selection_string);
 		g_free (localized_string);
@@ -1247,8 +1310,8 @@ selection_received (GtkWidget *widget,
 		if (selection_data->type != GDK_SELECTION_TYPE_STRING) {
 			html_engine_paste_text (GTK_HTML (widget)->engine, 
 						selection_data->data,
-						(guint) unicode_strlen (selection_data->data, 
-									selection_data->length));
+						g_utf8_strlen (selection_data->data, 
+							       selection_data->length));
 		} else {
 			gchar *utf8 = NULL;
 			
@@ -1257,7 +1320,7 @@ selection_received (GtkWidget *widget,
 						       (guint) selection_data->length);
 
 			html_engine_paste_text (GTK_HTML (widget)->engine, 
-						utf8, unicode_strlen (utf8, -1));
+						utf8, g_utf8_strlen (utf8, -1));
 
 			g_free (utf8);
 		}
@@ -1471,15 +1534,51 @@ init_properties (GtkHTMLClass *klass)
 }
 
 static gint
-focus (GtkContainer *c, GtkDirectionType direction)
+focus (GtkContainer *container, GtkDirectionType direction)
 {
+#if 1
 	gint rv;
-
-	GTK_WIDGET_UNSET_FLAGS (GTK_WIDGET (c), GTK_CAN_FOCUS);
-	rv = (*GTK_CONTAINER_CLASS (parent_class)->focus) (c, direction);
-	GTK_WIDGET_SET_FLAGS (GTK_WIDGET (c), GTK_CAN_FOCUS);
-
+	/* GTK_WIDGET_UNSET_FLAGS (GTK_WIDGET (container), GTK_CAN_FOCUS);*/
+	rv = (*GTK_CONTAINER_CLASS (parent_class)->focus) (container, direction);
+	/* GTK_WIDGET_SET_FLAGS (GTK_WIDGET (container), GTK_CAN_FOCUS);*/
+	html_engine_set_focus (GTK_HTML (container)->engine, rv);
 	return rv;
+#else
+	GtkWidget *focus_child;
+	GtkHTML *html;
+
+	g_return_val_if_fail (container != NULL, FALSE);
+	g_return_val_if_fail (GTK_IS_HTML (container), FALSE);
+
+	if (!GTK_WIDGET_IS_SENSITIVE (container))
+		return FALSE;
+
+	if (GTK_WIDGET_HAS_FOCUS (container))
+		return FALSE;
+	    
+	html = GTK_HTML (container);
+	focus_child = container->focus_child;
+	
+	switch (direction) {
+	case GTK_DIR_LEFT:
+	case GTK_DIR_RIGHT:
+		{
+		}
+	case GTK_DIR_DOWN:
+	case GTK_DIR_TAB_FORWARD:
+		{
+		}
+	case GTK_DIR_UP:
+	case GTK_DIR_TAB_BACKWARD:
+		{
+		}
+	default:
+		break;
+	}
+
+	gtk_container_set_focus_child (container, NULL);
+	return FALSE;
+#endif
 }
 
 typedef void (*GtkSignal_NONE__INT_INT_FLOAT) (GtkObject * object,
@@ -2148,8 +2247,8 @@ gtk_html_load_from_string  (GtkHTML *html, const gchar *str, gint len)
 	GtkHTMLStream *stream;
 
 	stream = gtk_html_begin (html);
-	gtk_html_write (html, stream, str, (len == -1) ? strlen (str) : len);
-	gtk_html_end (html, stream, GTK_HTML_STREAM_OK);
+	gtk_html_stream_write (stream, str, (len == -1) ? strlen (str) : len);
+	gtk_html_stream_close (stream, GTK_HTML_STREAM_OK);
 }
 
 
@@ -2340,6 +2439,7 @@ gtk_html_undo (GtkHTML *html)
 	g_return_if_fail (GTK_IS_HTML (html));
 
 	html_engine_undo (html->engine);
+	update_styles (html);
 }
 
 void
@@ -2349,6 +2449,7 @@ gtk_html_redo (GtkHTML *html)
 	g_return_if_fail (GTK_IS_HTML (html));
 
 	html_engine_redo (html->engine);
+	update_styles (html);
 }
 
 /* misc utils */
@@ -2466,9 +2567,14 @@ cursor_move (GtkHTML *html, GtkDirectionType dir_type, GtkHTMLCursorSkipType ski
 	if (!html_engine_get_editable (html->engine))
 		return;
 
-	if (html->engine->shift_selection)
+	if (html->engine->selection_mode) {
+		if (!html->engine->mark)
+			html_engine_set_mark (html->engine);
+	} else if (html->engine->shift_selection || html->engine->mark) {
 		html_engine_disable_selection (html->engine);
-
+		html_engine_edit_selection_updater_schedule (html->engine->selection_updater);
+		html->engine->shift_selection = FALSE;
+	}
 	switch (skip) {
 	case GTK_HTML_CURSOR_SKIP_ONE:
 		switch (dir_type) {
@@ -2556,69 +2662,148 @@ cursor_move (GtkHTML *html, GtkDirectionType dir_type, GtkHTMLCursorSkipType ski
 	gtk_html_edit_make_cursor_visible (html);
 }
 
-static void
+static gboolean
 move_selection (GtkHTML *html, GtkHTMLCommandType com_type)
 {
+	gboolean rv;
 	gint amount;
 
 	if (!html_engine_get_editable (html->engine))
-		return;
+		return FALSE;
 
 	html->engine->shift_selection = TRUE;
 	if (!html->engine->mark)
 		html_engine_set_mark (html->engine);
 	switch (com_type) {
 	case GTK_HTML_COMMAND_MODIFY_SELECTION_UP:
-		html_engine_move_cursor (html->engine, HTML_ENGINE_CURSOR_UP, 1);
+		rv = html_engine_move_cursor (html->engine, HTML_ENGINE_CURSOR_UP, 1) > 0 ? TRUE : FALSE;
 		break;
 	case GTK_HTML_COMMAND_MODIFY_SELECTION_DOWN:
-		html_engine_move_cursor (html->engine, HTML_ENGINE_CURSOR_DOWN, 1);
+		rv = html_engine_move_cursor (html->engine, HTML_ENGINE_CURSOR_DOWN, 1) > 0 ? TRUE : FALSE;
 		break;
 	case GTK_HTML_COMMAND_MODIFY_SELECTION_LEFT:
-		html_engine_move_cursor (html->engine, HTML_ENGINE_CURSOR_LEFT, 1);
+		rv = html_engine_move_cursor (html->engine, HTML_ENGINE_CURSOR_LEFT, 1) > 0 ? TRUE : FALSE;
 		break;
 	case GTK_HTML_COMMAND_MODIFY_SELECTION_RIGHT:
-		html_engine_move_cursor (html->engine, HTML_ENGINE_CURSOR_RIGHT, 1);
+		rv = html_engine_move_cursor (html->engine, HTML_ENGINE_CURSOR_RIGHT, 1) > 0 ? TRUE : FALSE;
 		break;
 	case GTK_HTML_COMMAND_MODIFY_SELECTION_BOL:
-		html_engine_beginning_of_line (html->engine);
+		rv = html_engine_beginning_of_line (html->engine);
 		break;
 	case GTK_HTML_COMMAND_MODIFY_SELECTION_EOL:
-		html_engine_end_of_line (html->engine);
+		rv = html_engine_end_of_line (html->engine);
 		break;
 	case GTK_HTML_COMMAND_MODIFY_SELECTION_BOD:
 		html_engine_beginning_of_document (html->engine);
+		rv = TRUE;
 		break;
 	case GTK_HTML_COMMAND_MODIFY_SELECTION_EOD:
 		html_engine_end_of_document (html->engine);
+		rv = TRUE;
+		break;
+	case GTK_HTML_COMMAND_MODIFY_SELECTION_PREV_WORD:
+		rv = html_engine_backward_word (html->engine);
+		break;
+	case GTK_HTML_COMMAND_MODIFY_SELECTION_NEXT_WORD:
+		rv = html_engine_forward_word (html->engine);
 		break;
 	case GTK_HTML_COMMAND_MODIFY_SELECTION_PAGEUP:
-		if ((amount = html_engine_scroll_up (html->engine, GTK_WIDGET (html)->allocation.height)) > 0)
+		if ((amount = html_engine_scroll_up (html->engine, GTK_WIDGET (html)->allocation.height)) > 0) {
 			scroll_by_amount (html, - amount);
+			rv = TRUE;
+		} else
+			rv = FALSE;
 		break;
 	case GTK_HTML_COMMAND_MODIFY_SELECTION_PAGEDOWN:
-		if ((amount = html_engine_scroll_down (html->engine, GTK_WIDGET (html)->allocation.height)) > 0)
+		if ((amount = html_engine_scroll_down (html->engine, GTK_WIDGET (html)->allocation.height)) > 0) {
 			scroll_by_amount (html, amount);
+			rv = TRUE;
+		} else
+			rv = FALSE;
 		break;
 	default:
 		g_warning ("invalid move_selection parameters\n");
+		rv = FALSE;
 	}
 
 	html->binding_handled = TRUE;
 	html->priv->update_styles = TRUE;
-	gtk_html_edit_make_cursor_visible (html);
+
+	return rv;
+}
+
+inline static void
+delete_one (HTMLEngine *e, gboolean forward)
+{
+	if (e->cursor->object && html_object_is_container (e->cursor->object)
+	    && ((forward && !e->cursor->offset) || (!forward && e->cursor->offset)))
+		html_engine_delete_container (e);
+	else
+		html_engine_delete_n (e, 1, forward);
+}
+
+inline static void
+inset_tab_or_indent_more_or_next_cell (GtkHTML *html)
+{
+	HTMLEngine *e = html->engine;
+	HTMLTableCell *cell;
+
+	cell = html_engine_get_table_cell (e);
+	if (cell) {
+		html_engine_hide_cursor (e);
+		html_cursor_end_of_line (e->cursor, e);
+		html_cursor_forward (e->cursor, e);
+		if (HTML_IS_TABLE (e->cursor->object)) {
+			html_cursor_backward (e->cursor, e);
+			html_engine_insert_table_row (e, TRUE);
+		}
+		html_engine_show_cursor (e);
+		gtk_html_edit_make_cursor_visible (html);
+	} else if (!html_engine_is_selection_active (e)
+	    && html_clueflow_tabs (HTML_CLUEFLOW (e->cursor->object->parent), e->painter))
+		html_engine_insert_text (e, "\t", 1);
+	else
+		gtk_html_modify_indent_by_delta (html, +1);
 }
 
 static void
+indent_less_or_prev_cell (GtkHTML *html)
+{
+	HTMLTableCell *cell;
+
+	cell = html_engine_get_table_cell (html->engine);
+	if (cell) {
+		HTMLEngine *e = html->engine;
+		html_engine_hide_cursor (e);
+		html_cursor_beginning_of_line (e->cursor, e);
+		html_cursor_backward (e->cursor, e);
+		html_engine_show_cursor (e);
+		gtk_html_edit_make_cursor_visible (html);
+	} else
+		gtk_html_modify_indent_by_delta (html, -1);
+
+}
+
+static gboolean
 command (GtkHTML *html, GtkHTMLCommandType com_type)
 {
 	HTMLEngine *e = html->engine;
+	gboolean rv = TRUE;
 
 	/* printf ("command %d %s\n", com_type, get_value_nick (com_type)); */
 	html->binding_handled = TRUE;
 
 	/* non-editable + editable commands */
 	switch (com_type) {
+	case GTK_HTML_COMMAND_ZOOM_IN:
+		gtk_html_zoom_in (html);
+		break;
+	case GTK_HTML_COMMAND_ZOOM_OUT:
+		gtk_html_zoom_out (html);
+		break;
+	case GTK_HTML_COMMAND_ZOOM_RESET:
+		gtk_html_zoom_reset (html);
+		break;
 	case GTK_HTML_COMMAND_SEARCH_INCREMENTAL_FORWARD:
 		gtk_html_isearch (html, TRUE);
 		break;
@@ -2636,7 +2821,7 @@ command (GtkHTML *html, GtkHTMLCommandType com_type)
 	}
 
 	if (!html_engine_get_editable (e) || html->binding_handled)
-		return;
+		return rv;
 
 	html->binding_handled = TRUE;
 	html->priv->update_styles = FALSE;
@@ -2673,13 +2858,13 @@ command (GtkHTML *html, GtkHTMLCommandType com_type)
 		    && e->mark->position != e->cursor->position)
 			html_engine_delete (e);
 		else
-			html_engine_delete_n (e, 1, TRUE);
+			delete_one (e, TRUE);
 		break;
 	case GTK_HTML_COMMAND_DELETE_BACK:
 		if (html_engine_is_selection_active (e))
 			html_engine_delete (e);
 		else
-			html_engine_delete_n (e, 1, FALSE);
+			delete_one (e, FALSE);
 		break;
 	case GTK_HTML_COMMAND_DELETE_BACK_OR_INDENT_DEC:
 		if (html_engine_is_selection_active (e))
@@ -2687,13 +2872,15 @@ command (GtkHTML *html, GtkHTMLCommandType com_type)
 		else if (html_engine_cursor_on_bop (e) && html_engine_get_indent (e) > 0)
 			gtk_html_modify_indent_by_delta (html, -1);
 		else
-			html_engine_delete_n (e, 1, FALSE);
+			delete_one (e, FALSE);
 		break;
-	case GTK_HTML_COMMAND_SET_MARK:
-		html_engine_set_mark (e);
+	case GTK_HTML_COMMAND_SELECTION_MODE:
+		e->selection_mode = TRUE;
 		break;
 	case GTK_HTML_COMMAND_DISABLE_SELECTION:
 		html_engine_disable_selection (e);
+		html_engine_edit_selection_updater_schedule (e->selection_updater);
+		e->selection_mode = FALSE;
 		break;
 	case GTK_HTML_COMMAND_BOLD_ON:
 		gtk_html_set_font_style (html, GTK_HTML_FONT_STYLE_MAX, GTK_HTML_FONT_STYLE_BOLD);
@@ -2771,11 +2958,29 @@ command (GtkHTML *html, GtkHTMLCommandType com_type)
 		gtk_html_set_indent (html, 0);
 		break;
 	case GTK_HTML_COMMAND_INDENT_INC:
-		gtk_html_modify_indent_by_delta (html, +1);
+			gtk_html_modify_indent_by_delta (html, +1);
+		break;
+	case GTK_HTML_COMMAND_INSERT_TAB:
+		if (!html_engine_is_selection_active (e)
+		    && html_clueflow_tabs (HTML_CLUEFLOW (e->cursor->object->parent), e->painter))
+			html_engine_insert_text (e, "\t", 1);
+		break;
+	case GTK_HTML_COMMAND_INSERT_TAB_OR_INDENT_MORE:
+		if (!html_engine_is_selection_active (e)
+		    && html_clueflow_tabs (HTML_CLUEFLOW (e->cursor->object->parent), e->painter))
+			html_engine_insert_text (e, "\t", 1);
+		else
+			gtk_html_modify_indent_by_delta (html, +1);
+		break;
+	case GTK_HTML_COMMAND_INSERT_TAB_OR_INDENT_MORE_OR_NEXT_CELL:
+		inset_tab_or_indent_more_or_next_cell (html);
 		break;
 	case GTK_HTML_COMMAND_INDENT_DEC:
 		gtk_html_modify_indent_by_delta (html, -1);
 		break;
+	case GTK_HTML_COMMAND_INDENT_DEC_OR_PREV_CELL:
+		indent_less_or_prev_cell (html);
+		break;
 	case GTK_HTML_COMMAND_INDENT_PARAGRAPH:
 		html_engine_indent_pre_paragraph (e);
 		break;
@@ -2831,7 +3036,9 @@ command (GtkHTML *html, GtkHTMLCommandType com_type)
 	case GTK_HTML_COMMAND_MODIFY_SELECTION_EOD:
 	case GTK_HTML_COMMAND_MODIFY_SELECTION_PAGEUP:
 	case GTK_HTML_COMMAND_MODIFY_SELECTION_PAGEDOWN:
-		move_selection (html, com_type);
+	case GTK_HTML_COMMAND_MODIFY_SELECTION_PREV_WORD:
+	case GTK_HTML_COMMAND_MODIFY_SELECTION_NEXT_WORD:
+		rv = move_selection (html, com_type);
 		break;
 	case GTK_HTML_COMMAND_SELECT_WORD:
 		gtk_html_select_word (html);
@@ -2906,23 +3113,38 @@ command (GtkHTML *html, GtkHTMLCommandType com_type)
 		html_engine_delete_table_row (e);
 		break;
 	case GTK_HTML_COMMAND_TABLE_BORDER_WIDTH_INC:
-		html_engine_table_set_border_width (e, 1, TRUE);
+		html_engine_table_set_border_width (e, html_engine_get_table (e), 1, TRUE);
 		break;
 	case GTK_HTML_COMMAND_TABLE_BORDER_WIDTH_DEC:
-		html_engine_table_set_border_width (e, -1, TRUE);
+		html_engine_table_set_border_width (e, html_engine_get_table (e), -1, TRUE);
 		break;
 	case GTK_HTML_COMMAND_TABLE_BORDER_WIDTH_ZERO:
-		html_engine_table_set_border_width (e, 0, FALSE);
+		html_engine_table_set_border_width (e, html_engine_get_table (e), 0, FALSE);
 		break;
 	case GTK_HTML_COMMAND_TEXT_SET_DEFAULT_COLOR:
 		html_engine_set_color (e, NULL);
 		break;
+	case GTK_HTML_COMMAND_CURSOR_BOD:
+		html_engine_beginning_of_document (e);		
+		break;
+	case GTK_HTML_COMMAND_CURSOR_EOD:
+		html_engine_end_of_document (e);
+		break;
+	case GTK_HTML_COMMAND_BLOCK_REDRAW:
+		html_engine_block_redraw (e);
+		break;
+	case GTK_HTML_COMMAND_UNBLOCK_REDRAW:
+		html_engine_unblock_redraw (e);
+		break;
+
 	default:
 		html->binding_handled = FALSE;
 	}
 
 	if (!html->binding_handled && html->editor_api)
 		html->binding_handled = (* html->editor_api->command) (html, com_type, html->editor_data);
+
+	return rv;
 }
 
 /*
@@ -3101,14 +3323,24 @@ load_keybindings (GtkHTMLClass *klass)
 	BCOM (0, ISO_Left_Tab, FOCUS_FORWARD);
 	BCOM (GDK_SHIFT_MASK, Tab, FOCUS_BACKWARD);
 	BCOM (GDK_SHIFT_MASK, ISO_Left_Tab, FOCUS_BACKWARD);
+
+	BCOM (GDK_CONTROL_MASK, equal, ZOOM_IN);
+	BCOM (GDK_CONTROL_MASK, KP_Add, ZOOM_IN);
+	BCOM (GDK_CONTROL_MASK, minus, ZOOM_OUT);
+	BCOM (GDK_CONTROL_MASK, KP_Subtract, ZOOM_OUT);
+	BCOM (GDK_CONTROL_MASK, 8, ZOOM_RESET);
+	BCOM (GDK_CONTROL_MASK, KP_Multiply, ZOOM_RESET);
+	BCOM (GDK_CONTROL_MASK, space, SELECTION_MODE);
+	BCOM (0, Escape, DISABLE_SELECTION);
 }
 
 void
-gtk_html_set_iframe_parent (GtkHTML *html, GtkWidget *parent)
+gtk_html_set_iframe_parent (GtkHTML *html, GtkWidget *parent, HTMLObject *frame)
 {
 	g_assert (GTK_IS_HTML (parent));
 
 	html->iframe_parent = parent;
+	html->frame = frame;
 	gtk_signal_emit (GTK_OBJECT (html_engine_get_top_html_engine (html->engine)->widget),
 			 signals [IFRAME_CREATED], html);
 }
@@ -3218,7 +3450,7 @@ get_value_nick (GtkHTMLCommandType com_type)
 void
 gtk_html_editor_event_command (GtkHTML *html, GtkHTMLCommandType com_type)
 {
-	if (html->editor_api) {
+	if (html->editor_api && !html->engine->block_events) {
 		GtkArg *args [1];
 
 		args [0] = gtk_arg_new (GTK_TYPE_STRING);
@@ -3230,14 +3462,13 @@ gtk_html_editor_event_command (GtkHTML *html, GtkHTMLCommandType com_type)
 	}
 }
 
-void
-gtk_html_editor_command (GtkHTML *html, const gchar *command_name)
+gboolean
+gtk_html_command (GtkHTML *html, const gchar *command_name)
 {
 	GtkEnumValue *val;
 
 	val = gtk_type_enum_find_value (GTK_TYPE_HTML_COMMAND, command_name);
-	if (val)
-		command (html, val->value);
+	return val ? command (html, val->value) : FALSE;
 }
 
 gboolean
@@ -3265,3 +3496,68 @@ gtk_html_build_with_gconf ()
 	return FALSE;
 #endif
 }
+
+void
+gtk_html_insert_html (GtkHTML *html, const gchar *html_src)
+{
+	GtkHTML *tmp;
+	GtkWidget *window, *sw;
+	HTMLObject *o;
+
+	tmp    = GTK_HTML (gtk_html_new_from_string (html_src, -1));
+	window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
+	sw     = gtk_scrolled_window_new (NULL, NULL);
+	gtk_container_add (GTK_CONTAINER (window), GTK_WIDGET (sw));
+	gtk_container_add (GTK_CONTAINER (sw), GTK_WIDGET (tmp));
+	gtk_widget_realize (GTK_WIDGET (tmp));
+	o                 = tmp->engine->clue;
+	tmp->engine->clue = NULL;
+	html_image_factory_move_images (html->engine->image_factory, tmp->engine->image_factory);
+	html_engine_insert_object (html->engine, o, html_object_get_recursive_length (o));
+	gtk_widget_destroy (window);
+}
+
+static void
+set_magnification (HTMLObject *o, HTMLEngine *e, gpointer data)
+{
+	if (HTML_IS_FRAME (o)) {
+		html_font_manager_set_magnification (&GTK_HTML (HTML_FRAME (o)->html)->engine->painter->font_manager,
+						     *(gdouble *) data);
+	} else if (HTML_IS_IFRAME (o)) {
+		html_font_manager_set_magnification (&GTK_HTML (HTML_IFRAME (o)->html)->engine->painter->font_manager,
+						     *(gdouble *) data);
+	}
+}
+
+void
+gtk_html_set_magnification (GtkHTML *html, gdouble magnification)
+{
+	if (magnification > 0.05 && magnification < 20.0
+	    && magnification * html->engine->painter->font_manager.var_size >= 4
+	    && magnification * html->engine->painter->font_manager.fix_size >= 4) {
+		html_object_forall (html->engine->clue, html->engine, set_magnification, &magnification);
+		html_font_manager_set_magnification (&html->engine->painter->font_manager, magnification);
+		html_object_change_set_down (html->engine->clue, HTML_CHANGE_ALL);
+		html_engine_schedule_update (html->engine);
+	}
+}
+
+#define MAG_STEP 1.1
+
+void
+gtk_html_zoom_in (GtkHTML *html)
+{
+	gtk_html_set_magnification (html, html->engine->painter->font_manager.magnification * MAG_STEP);
+}
+
+void
+gtk_html_zoom_out (GtkHTML *html)
+{
+	gtk_html_set_magnification (html, html->engine->painter->font_manager.magnification * (1.0/MAG_STEP));
+}
+
+void
+gtk_html_zoom_reset (GtkHTML *html)
+{
+	gtk_html_set_magnification (html, 1.0);
+}
