@@ -87,7 +87,7 @@ op_helper (HTMLObject *self, HTMLEngine *e, GList *from, GList *to, GList *left,
 {
 	HTMLClue *clue = HTML_CLUE (self);
 	HTMLObject *cc;
-	HTMLObject *o, *last, *cnext;
+	HTMLObject *o, *last, *cnext, *child;
 
 	cc   = html_object_dup (self);
 	o    = (from) ? HTML_OBJECT (from->data) : clue->head;
@@ -98,14 +98,15 @@ op_helper (HTMLObject *self, HTMLEngine *e, GList *from, GList *to, GList *left,
 
 	while (1) {
 		cnext = html_object_next_not_slave (o);
-		html_clue_append (HTML_CLUE (cc), cut
-				  ? html_object_op_cut (o, e,
-							html_object_get_bound_list (o, from),
-							html_object_get_bound_list (o, to),
-							left ? left->next : NULL, right ? right->next : NULL, len)
-				  : html_object_op_copy (o, e,
-							 html_object_get_bound_list (o, from),
-							 html_object_get_bound_list (o, to), len));
+		child = cut ? html_object_op_cut (o, e,
+					      html_object_get_bound_list (o, from),
+					      html_object_get_bound_list (o, to),
+					      left ? left->next : NULL, right ? right->next : NULL, len)
+			: html_object_op_copy (o, e,
+					       html_object_get_bound_list (o, from),
+					       html_object_get_bound_list (o, to), len);
+		if (child)
+			html_clue_append (HTML_CLUE (cc), child);
 
 		if (o == last)
 			break;
@@ -134,7 +135,7 @@ op_cut (HTMLObject *self, HTMLEngine *e, GList *from, GList *to, GList *left, GL
 			html_object_remove_child (self->parent, self);
 		html_object_destroy (self);
 	} else
-		html_object_change_set (self, HTML_CHANGE_ALL);
+		html_object_change_set (self, HTML_CHANGE_ALL_CALC);
 
 	return rv;
 }
@@ -151,9 +152,7 @@ merge (HTMLObject *self, HTMLObject *with, HTMLEngine *e, GList *left, GList *ri
 	clue2->head = NULL;
 	clue2->tail = NULL;
 
-	printf ("merge clues\n");
-
-	html_object_change_set (self, HTML_CHANGE_ALL);
+	html_object_change_set (self, HTML_CHANGE_ALL_CALC);
 	return TRUE;
 }
 
@@ -227,7 +226,7 @@ set_max_ascent (HTMLObject *o,
 	HTMLClue *clue = HTML_CLUE (o);
 	HTMLObject *obj;
 
-	if (clue->valign == HTML_VALIGN_CENTER) {
+	if (clue->valign == HTML_VALIGN_MIDDLE) {
 		for (obj = HTML_CLUE (o)->head; obj != 0; obj = obj->next) {
 			obj->y = obj->y + ((a - o->ascent) / 2);
 		}
@@ -248,7 +247,7 @@ set_max_descent (HTMLObject *o, HTMLPainter *painter, gint d)
 	HTMLClue *clue = HTML_CLUE (o);
 	HTMLObject *obj;
 	
-	if (clue->valign == HTML_VALIGN_CENTER) {
+	if (clue->valign == HTML_VALIGN_MIDDLE) {
 		for (obj = clue->head; obj != 0; obj = obj->next) {
 			obj->y = obj->y + ((d - o->descent) / 2);
 		}
@@ -489,7 +488,8 @@ get_right_clear (HTMLClue *o, gint y)
 }
 
 static void
-find_free_area (HTMLClue *clue, gint y,
+find_free_area (HTMLClue *clue,
+		gint y,
 		gint width, gint height,
 		gint indent, gint *y_pos,
 		gint *lmargin, gint *rmargin)
@@ -658,13 +658,12 @@ html_clue_get_right_clear (HTMLClue *clue, gint y)
 }
 
 void
-html_clue_find_free_area (HTMLClue *clue, gint y,
+html_clue_find_free_area (HTMLClue *clue,
+			  gint y,
 			  gint width, gint height, gint indent, gint *y_pos,
 			  gint *lmargin, gint *rmargin)
 {
-	(* HC_CLASS (clue)->find_free_area) (clue, y, width, height,
-					     indent, y_pos,
-					     lmargin, rmargin);
+	(* HC_CLASS (clue)->find_free_area) (clue, y, width, height, indent, y_pos, lmargin, rmargin);
 }
 
 void
