@@ -23,7 +23,9 @@
 */
 
 #include <config.h>
+#include <stdio.h>
 #include <string.h>
+
 #include "gtkhtmldebug.h"
 #include "htmlcolor.h"
 #include "htmlcolorset.h"
@@ -133,7 +135,7 @@ copy_sized (HTMLObject *self, HTMLObject *dest, gint rows, gint cols)
 	for (r = 0; r < rows; r++)
 		d->cells [r] = g_new0 (HTMLTableCell *, cols);
 
-	dest->change = HTML_CHANGE_ALL;
+	dest->change = HTML_CHANGE_ALL_CALC;
 }
 
 static void
@@ -182,7 +184,6 @@ op_copy (HTMLObject *self, HTMLEngine *e, GList *from, GList *to, guint *len)
 
 	printf ("cols: %d rows: %d\n", cols, rows);
 
-	*len = 0;
 	for (r = 0; r < rows; r++)
 		for (c = 0; c < cols; c++) {
 			HTMLTableCell *cell = t->cells [start->row + r][c + start_col];
@@ -201,7 +202,7 @@ op_copy (HTMLObject *self, HTMLEngine *e, GList *from, GList *to, guint *len)
 			} else
 				nt->cells [r][c] = nt->cells [cell->row - start->row][cell->col - start_col];
 		}
-
+	(*len) ++;
 	if (end->col - start_col < cols - 1)
 		do_cspan (nt, nt->totalRows - 1, end->col - start_col, nt->cells [nt->totalRows - 1][end->col - start_col]);
 
@@ -265,15 +266,14 @@ cut_partial (HTMLObject *self, HTMLEngine *e, GList *from, GList *to, GList *lef
 
 	printf ("partial cut\n");
 
-	start = HTML_TABLE_CELL (from ? from->data : html_object_head (self));
-	end   = HTML_TABLE_CELL (to   ? to->data   : html_object_tail (self));
+	start = HTML_TABLE_CELL (from && from->next ? from->data : html_object_head (self));
+	end   = HTML_TABLE_CELL (to   && to->next   ? to->data   : html_object_tail (self));
 
 	start_row = start->row;
 	start_col = start->col;
 	end_row   = end->row;
 	end_col   = end->col;
 
-	*len = 0;
 	t    = HTML_TABLE (self);
 	rows = end_row - start_row + 1;
 	cols = end_row == start_row ? end_col - start_col + 1 : t->totalCols;
@@ -302,11 +302,14 @@ cut_partial (HTMLObject *self, HTMLEngine *e, GList *from, GList *to, GList *lef
 				html_table_cell_set_position (cell_cut, row, col);
 			}
 		}
-
-	shrink = start_row == 0 && end_row == t->totalRows - 1;
+	(*len) ++;
+	shrink = start_row == 0 && end_row == t->totalRows - 1 && start_col != end_col;
 	/* move remaining cells in old table */
-	if (start_col >= end_col)
+	if (start_col > end_col)
 		start_row ++;
+	else if (start_col == end_col && start != end) {
+		html_object_merge_down (HTML_OBJECT (start), HTML_OBJECT (end), e);
+	}
 	if (start_row != end_row)
 		for (r = end_row; r < t->totalRows; r ++)
 			for (c = 0; c < t->totalCols; c ++) {
@@ -326,6 +329,7 @@ cut_partial (HTMLObject *self, HTMLEngine *e, GList *from, GList *to, GList *lef
 		t->totalCols -= end_col - start_col - 1;
 	t->totalRows -= end_row - start_row;
 	printf ("removed partial table len: %d\n", *len);
+	gtk_html_debug_dump_tree_simple (rv, 0);
 
 	return rv;
 }
@@ -412,8 +416,8 @@ split (HTMLObject *self, HTMLEngine *e, HTMLObject *child, gint offset, gint lev
 	*left  = g_list_prepend (*left, self);
 	*right = g_list_prepend (*right, dup);
 
-	html_object_change_set (self, HTML_CHANGE_ALL);
-	html_object_change_set (dup,  HTML_CHANGE_ALL);
+	html_object_change_set (self, HTML_CHANGE_ALL_CALC);
+	html_object_change_set (dup,  HTML_CHANGE_ALL_CALC);
 
 	printf ("after split\n");
 	printf ("-- self --\n");
@@ -510,7 +514,7 @@ merge (HTMLObject *self, HTMLObject *with, HTMLEngine *e, GList *left, GList *ri
 			}
 		}
 
-	html_object_change_set (self, HTML_CHANGE_ALL);
+	html_object_change_set (self, HTML_CHANGE_ALL_CALC);
 
 	printf ("after merge\n");
 	printf ("-- self --\n");
@@ -630,7 +634,6 @@ cell_end_row (HTMLTable *table, HTMLTableCell *cell)
 
 #define ARR(i) (g_array_index (array, gint, i))
 #define LL (unsigned long long)
-// #define LL (gdouble)
 
 static gboolean
 calc_column_width_step (HTMLTable *table, HTMLPainter *painter, GArray *array, gint *sizes,
@@ -653,8 +656,8 @@ calc_column_width_step (HTMLTable *table, HTMLPainter *painter, GArray *array, g
 			if (cspan != span)
 				continue;
 
-			col_width  = (*calc_fn) (HTML_OBJECT (cell), painter) + pixel_size * table->padding * 2
-				- (span - 1) * (table->padding * 2 + table->spacing + border_extra) * pixel_size;
+			col_width  = (*calc_fn) (HTML_OBJECT (cell), painter)
+				- (span - 1) * (table->spacing + border_extra) * pixel_size;
 			if (col_width <= 0)
 				continue;
 			span_width = ARR (cell->col + span) - ARR (cell->col);
@@ -979,7 +982,7 @@ draw (HTMLObject *o,
 					 ROW_HEIGHT (table, table->totalRows) +
 					 pixel_size * table->border, GTK_HTML_ETCH_OUT,
 					 pixel_size * table->border);
-		
+
 		/* Draw borders around each cell */
 		for (r = start_row; r <= end_row; r++) {
 			for (c = start_col; c <= end_col; c++) {
@@ -1055,7 +1058,7 @@ calc_percentage_step (HTMLTable *table, gint *col_percent, gint *span_percent, g
 			if (!cell || cell->col != c || cell->row != r)
 				continue;
 
-			if (HTML_OBJECT (cell)->flags & HTML_OBJECT_FLAG_FIXEDWIDTH || HTML_OBJECT (cell)->percent <= 0)
+			if (HTML_OBJECT (cell)->flags & HTML_OBJECT_FLAG_FIXEDWIDTH || !cell->percent_width)
 				continue;
 
 			cspan = MIN (cell->cspan, table->totalCols - cell->col);
@@ -1065,7 +1068,7 @@ calc_percentage_step (HTMLTable *table, gint *col_percent, gint *span_percent, g
 				continue;
 
 			cl = cell_end_col (table, cell);
-			if (col_percent [cl] - col_percent [c] < HTML_OBJECT (cell)->percent) {
+			if (col_percent [cl] - col_percent [c] < cell->fixed_width) {
 				gint cp, part, added, pleft, not_percented, np;
 				part = 0;
 				not_percented = 0;
@@ -1075,7 +1078,7 @@ calc_percentage_step (HTMLTable *table, gint *col_percent, gint *span_percent, g
 
 				np    = 1;
 				added = 0;
-				pleft = HTML_OBJECT (cell)->percent - (col_percent [cl] - col_percent [c]);
+				pleft = cell->fixed_width - (col_percent [cl] - col_percent [c]);
 				for (cp = 0; cp < span; cp++) {
 					if (not_percented) {
 						if (!PERC (c + cp)) {
@@ -1372,7 +1375,6 @@ html_table_set_cells_max_width (HTMLTable *table, HTMLPainter *painter, gint *ma
 				size = max_size [c] + (cell->col != c ? size : 0);
 				if (cell_end_col (table, cell) - 1 == c && cell->row == r)
 					html_object_set_max_width (HTML_OBJECT (cell), painter, size
-								   - pixel_size * table->padding * 2
 								   + pixel_size * (table->spacing + border_extra) * CSPAN);
 			}
 		}
@@ -1570,7 +1572,7 @@ check_point (HTMLObject *self,
 		for (c=0; c<table->totalCols; c++)
 			if (table->cells [start_row][c])
 				break;
-		if (table->cells [start_row][c]) {
+		if (c < table->totalCols && table->cells [start_row][c]) {
 			cell = table->cells [start_row][c];
 			if (x < HTML_OBJECT (cell)->x || y < HTML_OBJECT (cell)->y - HTML_OBJECT (cell)->ascent) {
 				obj = html_object_check_point (HTML_OBJECT (cell), painter,
@@ -1586,7 +1588,7 @@ check_point (HTMLObject *self,
 		for (c=table->totalCols - 1; c >= 0; c--)
 			if (table->cells [start_row][c])
 				break;
-		if (table->cells [start_row][c]) {
+		if (c >=0 && table->cells [start_row][c]) {
 			cell = table->cells [start_row][c];
 			if (x > HTML_OBJECT (cell)->x + HTML_OBJECT (cell)->width - 1
 			    || y > HTML_OBJECT (cell)->y + HTML_OBJECT (cell)->descent - 1) {
@@ -2096,4 +2098,14 @@ html_table_end_row (HTMLTable *table)
 void
 html_table_end_table (HTMLTable *table)
 {
+	gint r, c;
+
+	for (r = 0; r < table->totalRows; r ++)
+		for (c = 0; c < table->totalCols; c ++)
+			if (table->cells [r][c] && HTML_CLUE (table->cells [r][c])->head == NULL) {
+				HTMLTableCell *cell = table->cells [r][c];
+
+				remove_cell (table, cell);
+				html_object_destroy (HTML_OBJECT (cell));
+			}
 }
